import os
import random
from PIL import Image
from PyQt6 import uic
from PyQt6.QtCore import QRect, pyqtSignal, Qt
from PyQt6.QtGui import QShortcut, QKeySequence
from PyQt6.uic.exceptions import UIFileException
from aihandler.settings import MAX_SEED, MODELS, MessageCode
from airunner.windows.video import VideoPopup
from airunner.mixins.lora_mixin import LoraMixin
from PIL import PngImagePlugin


class GeneratorMixin(LoraMixin):
    generate_signal = pyqtSignal(dict)
    deterministic_generation_window = None
    deterministic_images = []
    deterministic_data = None

    @property
    def deterministic_var(self):
        return self.settings.deterministic

    @property
    def deterministic(self):
        return self.deterministic_var.get()

    @deterministic.setter
    def deterministic(self, val):
        self.deterministic_var.set(val == True)

    @property
    def available_models(self):
        return MODELS

    @property
    def model_base_path(self):
        return self.settings.model_base_path.get()

    @model_base_path.setter
    def model_base_path(self, val):
        self.settings.model_base_path.set(val)

    @property
    def working_width(self):
        return int(self.settings_manager.settings.working_width.get())

    @working_width.setter
    def working_width(self, val):
        self.settings_manager.settings.working_width.set(val)
        self.set_final_size_label()
        self.canvas.update()

    @property
    def working_height(self):
        return int(self.settings_manager.settings.working_height.get())

    @working_height.setter
    def working_height(self, val):
        self.settings_manager.settings.working_height.set(val)
        self.set_final_size_label()
        self.canvas.update()

    @property
    def steps_var(self):
        return self.settings.steps

    @property
    def steps(self):
        return self.steps_var.get()

    @steps.setter
    def steps(self, val):
        self.steps_var.set(val)

    @property
    def ddim_eta(self):
        return self.settings.ddim_eta.get()

    @ddim_eta.setter
    def ddim_eta(self, val):
        self.settings.ddim_eta.set(val)

    @property
    def prompt_var(self):
        return self.settings.prompt

    @property
    def prompt(self):
        return self.prompt_var.get()

    @prompt.setter
    def prompt(self, val):
        self.prompt_var.set(val)

    @property
    def negative_prompt_var(self):
        return self.settings.negative_prompt

    @property
    def negative_prompt(self):
        return self.negative_prompt_var.get()

    @negative_prompt.setter
    def negative_prompt(self, val):
        self.negative_prompt_var.set(val)

    @property
    def scale_var(self):
        return self.settings.scale

    @property
    def scale(self):
        return self.scale_var.get()

    @scale.setter
    def scale(self, val):
        self.scale_var.set(val)

    @property
    def image_scale_var(self):
        return self.settings.image_guidance_scale

    @property
    def image_scale(self):
        return self.image_scale_var.get()

    @image_scale.setter
    def image_scale(self, val):
        self.image_scale_var.set(val)

    @property
    def strength_var(self):
        return self.settings.strength

    @property
    def strength(self):
        return self.strength_var.get()

    @strength.setter
    def strength(self, val):
        self.strength_var.set(val)

    @property
    def zeroshot_var(self):
        return self.settings.zeroshot

    @property
    def zeroshot(self):
        return self.zeroshot_var.get()

    @zeroshot.setter
    def zeroshot(self, val):
        self.zeroshot_var.set(val)

    @property
    def enable_controlnet_var(self):
        return self.settings.enable_controlnet

    @property
    def enable_controlnet(self):
        return self.enable_controlnet_var.get()

    @enable_controlnet.setter
    def enable_controlnet(self, val):
        self.enable_controlnet_var.set(val)

    @property
    def controlnet_var(self):
        return self.settings.controlnet_var

    @property
    def controlnet(self):
        controlnet = self.controlnet_var.get()
        if controlnet == "":
            return None
        return controlnet

    @controlnet.setter
    def controlnet(self, val):
        self.controlnet_var.set(val)

    @property
    def use_prompt_builder_checkbox_var(self):
        return self.settings.use_prompt_builder_checkbox

    @property
    def use_prompt_builder_checkbox(self):
        val = self.use_prompt_builder_checkbox_var.get()
        return val == 2 or val == True

    @use_prompt_builder_checkbox.setter
    def use_prompt_builder_checkbox(self, val):
        self.use_prompt_builder_checkbox_var.set(val == True or val == 2)

    @property
    def controlnet_scale_var(self):
        return self.settings.controlnet_guidance_scale

    @property
    def controlnet_guidance_scale(self):
        return self.controlnet_scale_var.get()

    @controlnet_guidance_scale.setter
    def controlnet_guidance_scale(self, val):
        self.controlnet_scale_var.set(val)

    @property
    def seed_var(self):
        return self.settings.seed

    @property
    def seed(self):
        return self.seed_var.get()

    @seed.setter
    def seed(self, val):
        self.seed_var.set(val)

    @property
    def random_seed_var(self):
        return self.settings.random_seed

    @property
    def random_seed(self):
        return self.random_seed_var.get()

    @random_seed.setter
    def random_seed(self, val):
        self.random_seed_var.set(val)

    @property
    def samples_var(self):
        return self.settings.n_samples

    @property
    def samples(self):
        return self.samples_var.get()

    @samples.setter
    def samples(self, val):
        self.samples_var.set(val)

    @property
    def model_var(self):
        return self.settings.model_var

    @property
    def model(self):
        return self.model_var.get()

    @model.setter
    def model(self, val):
        self.model_var.set(val)

    @property
    def scheduler_var(self):
        return self.settings.scheduler_var

    @property
    def scheduler(self):
        return self.scheduler_var.get()

    @scheduler.setter
    def scheduler(self, val):
        self.scheduler_var.set(val)

    @property
    def downscale_amount_var(self):
        return self.settings.downscale_amount

    @property
    def downscale_amount(self):
        return self.downscale_amount_var.get()

    @downscale_amount.setter
    def downscale_amount(self, val):
        self.downscale_amount_var.set(val)

    @property
    def do_upscale_by_active_grid_var(self):
        return self.settings.do_upscale_by_active_grid

    @property
    def do_upscale_by_active_grid(self):
        return self.do_upscale_by_active_grid_var.get()

    @do_upscale_by_active_grid.setter
    def do_upscale_by_active_grid(self, val):
        self.do_upscale_by_active_grid_var.set(val)

    @property
    def do_upscale_full_image_var(self):
        return self.settings.do_upscale_full_image

    @property
    def do_upscale_full_image(self):
        return self.do_upscale_full_image_var.get()

    @do_upscale_full_image.setter
    def do_upscale_full_image(self, val):
        self.do_upscale_full_image_var.set(val)

    @property
    def image_to_new_layer(self):
        return self.settings_manager.settings.image_to_new_layer.get()

    def update_prompt(self, prompt):
        self.generator_tab_widget.set_prompt(prompt)

    def update_negative_prompt(self, prompt):
        self.generator_tab_widget.set_negative_prompt(prompt)

    def initialize(self):
        self.tool_menu_widget.initialize()
        self.initialize_lora()

        # listen to F5 keypress and call self.generate_callback
        self.generate_shortcut = QShortcut(QKeySequence(Qt.Key.Key_F5), self)
        self.generate_shortcut.activated.connect(self.generate_callback)

    def update_controlnet(self, tab, index):
        controlnet = self.tabs[tab].controlnet_dropdown.itemText(index)
        controlnet = controlnet.lower()
        if controlnet == "":
            controlnet = None
        self.settings.controlnet_var.set(controlnet)

    def set_final_size_label(self, tab=None):
        if tab is None:
            tab = self.tabs[self.current_section]

        """
        Early return to hack aronud final size for now
        """
        return

        image = self.canvas.current_layer.image_data.image
        if image:
            if self.do_upscale_by_active_grid:
                width = self.settings_manager.settings.working_width.get()
                height = self.settings_manager.settings.working_height.get()
            else:
                width = image.width
                height = image.height
        else:
            width = 0
            height = 0

        if self.downscale_amount > 0:
            width = width // self.downscale_amount
            height = height // self.downscale_amount

        if self.current_section == "upscale":
            width = width * 2
            height = height * 2
        elif self.current_section == "superresolution":
            width = width * 4
            height = height * 4

        # set final_size label text
        tab.final_size.setText(f"{width}x{height}")

    def handle_upscale_full_image_change(self, value, tab):
        self.do_upscale_full_image = value
        self.do_upscale_by_active_grid = value == False
        self.set_final_size_label(tab)

    def handle_upscale_active_grid_change(self, value, tab):
        self.do_upscale_by_active_grid = value
        self.do_upscale_full_image = value == False
        self.set_final_size_label(tab)

    def interrupt(self):
        print("Interrupting...")
        self.client.sd_runner.cancel()

    def reset_settings(self):
        self.settings_manager.reset_settings_to_default()
        for tab_name in self.tabs.keys():
            tab = self.tabs[tab_name]
            self.set_default_values(tab_name, tab)
        self.canvas.update()

    def text_changed(self, tab):
        try:
            val = int(tab.seed.toPlainText())
            self.seed = val
        except ValueError:
            pass

    def handle_random_checkbox_change(self, val, _tab):
        if val == 2:
            self.random_seed = True
        else:
            self.random_seed = False
        _tab.seed.setEnabled(not self.random_seed)

    def video_handler(self, data):
        filename = data["video_filename"]
        VideoPopup(settings_manager=self.settings_manager, file_path=filename)

    def prepare_metadata(self, data):
        if not self.settings_manager.settings.export_metadata.get() or \
                self.settings_manager.settings.image_export_type.get() != "png":
            return None
        metadata = PngImagePlugin.PngInfo()
        options = data["options"]
        action = data["action"]
        metadata.add_text("action", action)
        if self.settings_manager.settings.image_export_metadata_prompt.get() is True:
            metadata.add_text("prompt", options[f'{action}_prompt'])
        if self.settings_manager.settings.image_export_metadata_negative_prompt.get() is True:
            metadata.add_text("negative_prompt", options[f'{action}_negative_prompt'])
        if self.settings_manager.settings.image_export_metadata_scale.get() is True:
            metadata.add_text("scale", str(options[f"{action}_scale"]))
        if self.settings_manager.settings.image_export_metadata_seed.get() is True:
            metadata.add_text("seed", str(options[f"{action}_seed"]))
        if self.settings_manager.settings.image_export_metadata_steps.get() is True:
            metadata.add_text("steps", str(options[f"{action}_steps"]))
        if self.settings_manager.settings.image_export_metadata_ddim_eta.get() is True:
            metadata.add_text("ddim_eta", str(options[f"{action}_ddim_eta"]))
        if self.settings_manager.settings.image_export_metadata_iterations.get() is True:
            metadata.add_text("n_iter", str(options[f"{action}_n_iter"]))
        if self.settings_manager.settings.image_export_metadata_samples.get() is True:
            metadata.add_text("n_samples", str(options[f"{action}_n_samples"]))
        if self.settings_manager.settings.image_export_metadata_model.get() is True:
            metadata.add_text("model", str(options[f"{action}_model"]))
        if self.settings_manager.settings.image_export_metadata_model_branch.get() is True:
            metadata.add_text("model_branch", str(options[f"{action}_model_branch"]))
        if self.settings_manager.settings.image_export_metadata_scheduler.get() is True:
            metadata.add_text("scheduler", str(options[f"{action}_scheduler"]))
        return metadata

    def auto_export_image(self, image, data):
        """
        Export image along with stats to image_path
        :return:
        """
        if data["action"] == "txt2vid":
            return
        base_path = self.settings_manager.settings.model_base_path.get()
        image_path = self.settings_manager.settings.image_path.get()
        image_path = "images" if image_path == "" else image_path
        path = os.path.join(base_path, image_path) if image_path == "images" else image_path
        if not os.path.exists(path):
            os.makedirs(path)
        # check for existing files, if they exist, increment the filename. filename should be in the format
        # <action>_<seed>_<N>.png
        extension = f".{self.settings_manager.settings.image_export_type.get()}"
        filename = data["action"] + "_" + str(self.seed)
        if os.path.exists(os.path.join(path, filename + extension)):
            i = 1
            while os.path.exists(os.path.join(path, filename + "_" + str(i) + extension)):
                i += 1
            filename = filename + "_" + str(i)
        metadata = self.prepare_metadata(data)
        if metadata:
            image.save(os.path.join(path, filename + extension), pnginfo=metadata)
        else:
            image.save(os.path.join(path, filename + extension))

    def generate_callback(self):
        self.generate(True)

    def prep_video(self):
        pass

    @property
    def active_rect(self):
        rect = QRect(
            self.canvas.active_grid_area_rect.x(),
            self.canvas.active_grid_area_rect.y(),
            self.canvas.active_grid_area_rect.width(),
            self.canvas.active_grid_area_rect.height(),
        )
        rect.translate(-self.canvas.pos_x, -self.canvas.pos_y)

        return rect

    def tab_has_embeddings(self, tab):
        return tab not in ["upscale", "superresolution", "txt2vid"]

    def new_batch(self, index, image, data):
        """
        Generate a batch of images using deterministic geneartion based on a previous deterministic generation
        batch. The previous seed that was chosen should be re-used with the index added to it to generate the new
        batch of images.
        :return:
        """
        action = data["action"]
        if not data["options"]["deterministic_seed"]:
            data["options"][f"{action}_seed"] = data["options"][f"{action}_seed"] + index
        self.deterministic_data = data
        self.deterministic_index = index
        self.generate(image)
        self.deterministic_data = None
        self.deterministic_images = None

    def generate(self, image=None):
        if self.current_section in ("upscale", "superresolution") and self.do_upscale_full_image:
            image_data = self.canvas.current_layer.image_data
            image = image_data.image if image_data else None
            if image is None:
                self.message_var.emit({
                    "code": MessageCode.ERROR,
                    "message": "No image to upscale",
                })
                return
            downscale_amount = self.downscale_amount
            if downscale_amount > 0:
                # downscale the image first
                image = image.resize(
                    (
                        int(image.width // downscale_amount),
                        int(image.height // downscale_amount),
                    ),
                    Image.BICUBIC,
                )
            self.requested_image = image
            self.start_progress_bar()
            self.do_generate({
                "mask": image.convert("RGB"),
                "image": image.convert("RGB"),
                "location": image_data.position
            })
        elif self.use_pixels:
            self.requested_image = image
            self.start_progress_bar()
            image_data = self.canvas.current_layer.image_data
            image = image_data.image if image_data else None

            if image is None:
                # create a transparent image the size of self.canvas.active_grid_area_rect
                width = self.settings_manager.settings.working_width.get()
                height = self.settings_manager.settings.working_height.get()
                image = Image.new("RGBA", (int(width), int(height)), (0, 0, 0, 0))

            img = image.copy().convert("RGBA")
            new_image = Image.new(
                "RGBA",
                (self.settings.working_width.get(), self.settings.working_height.get()),
                (0, 0, 0))

            cropped_outpaint_box_rect = self.active_rect
            crop_location = (
                cropped_outpaint_box_rect.x() - self.canvas.image_pivot_point.x(),
                cropped_outpaint_box_rect.y() - self.canvas.image_pivot_point.y(),
                cropped_outpaint_box_rect.width() - self.canvas.image_pivot_point.x(),
                cropped_outpaint_box_rect.height() - self.canvas.image_pivot_point.y()
            )
            new_image.paste(img.crop(crop_location), (0, 0))
            # save new_image to disc
            mask = Image.new("RGB", (new_image.width, new_image.height), (255, 255, 255))
            for x in range(new_image.width):
                for y in range(new_image.height):
                    try:
                        if new_image.getpixel((x, y))[3] != 0:
                            mask.putpixel((x, y), (0, 0, 0))
                    except IndexError:
                        pass

            # convert image to rgb
            image = new_image.convert("RGB")
            self.do_generate({
                "mask": mask,
                "image": image,
                "location": self.canvas.active_grid_area_rect
            })
        elif self.action == "vid2vid":
            images = self.prep_video()
            self.do_generate({
                "images": images
            })
        else:
            self.do_generate()

    def start_progress_bar(self):
        # progressBar: QProgressBar = self.tabs[section].progressBar
        # progressBar.setRange(0, 0)
        self.generator_tab_widget.start_progress_bar(
            self.currentTabSection, self.current_section)

    def set_seed(self):
        """
        Set the seed - either set to random, deterministic or keep current, then display the seed in the UI.
        :return:
        """
        if self.deterministic_data:
            action = self.deterministic_data["action"]
            self.seed = self.deterministic_data["options"][f"{action}_seed"]
        elif self.random_seed:
            seed = random.randint(0, MAX_SEED)
            self.seed = seed
        self.generator_tab_widget.data[self.currentTabSection][self.current_section]["seed"].setText(str(self.seed))

    def get_memory_options(self):
        return {
            "use_last_channels": self.settings_manager.settings.use_last_channels.get(),
            "use_enable_sequential_cpu_offload": self.settings_manager.settings.use_enable_sequential_cpu_offload.get(),
            "enable_model_cpu_offload": self.settings_manager.settings.enable_model_cpu_offload.get(),
            "use_attention_slicing": self.settings_manager.settings.use_attention_slicing.get(),
            "use_tf32": self.settings_manager.settings.use_tf32.get(),
            "use_cudnn_benchmark": self.settings_manager.settings.use_cudnn_benchmark.get(),
            "use_enable_vae_slicing": self.settings_manager.settings.use_enable_vae_slicing.get(),
            "use_accelerated_transformers": self.settings_manager.settings.use_accelerated_transformers.get(),
            "use_torch_compile": self.settings_manager.settings.use_torch_compile.get(),
            "use_tiled_vae": self.settings_manager.settings.use_tiled_vae.get(),
        }

    def do_deterministic_generation(self, extra_options):
        action = self.deterministic_data["action"]
        options = self.deterministic_data["options"]
        options[f"{action}_prompt"] = self.deterministic_data[f"{action}_prompt"][self.deterministic_index]
        memory_options = self.get_memory_options()
        data = {
            "action": action,
            "options": {
                **options,
                **extra_options,
                **memory_options,
                "deterministic_generation": True,
                "deterministic_seed": True,
            }
        }
        self.client.message = data

    def do_generate(self, extra_options=None):
        if not extra_options:
            extra_options = {}

        self.set_seed()

        if self.deterministic_data and self.deterministic:
            return self.do_deterministic_generation(extra_options)

        # self.start_progress_bar(self.current_section)

        action = self.current_section
        tab = self.tabs[action]

        self.settings.seed.set(self.seed)

        if action in ("txt2img", "img2img", "pix2pix", "depth2img", "txt2vid"):
            samples = self.samples
        else:
            samples = 1

        prompt = self.prompt
        negative_prompt = self.negative_prompt

        # set the model data
        model = self.model
        model_branch = None
        section_name = action
        if section_name in ["txt2img", "img2img"]:
            section_name = "generate"

        section_name = f"{self.currentTabSection}_{section_name}"

        if model in self.available_models[section_name]:
            model_path = self.available_models[section_name][model]["path"]
            model_branch = self.available_models[section_name][model].get("branch", "main")
        elif model not in self.models:
            model_names = list(self.available_models[section_name].keys())
            model = model_names[0]
            model_path = self.available_models[section_name][model]["path"]
            model_branch = self.available_models[section_name][model].get("branch", "main")
        else:
            path = self.settings_manager.settings.model_base_path.get()
            if action == "depth2img":
                path = self.settings_manager.settings.depth2img_model_path.get()
            elif action == "pix2pix":
                path = self.settings_manager.settings.pix2pix_model_path.get()
            elif action == "outpaint":
                path = self.settings_manager.settings.outpaint_model_path.get()
            elif action == "upscale":
                path = self.settings_manager.settings.upscale_model_path.get()
            model_path = os.path.join(path, model)

        # get controlnet_dropdown from active tab
        options = {
            f"{action}_prompt": prompt,
            f"{action}_negative_prompt": negative_prompt,
            f"{action}_steps": self.steps,
            f"{action}_ddim_eta": self.ddim_eta,  # only applies to ddim scheduler
            f"{action}_n_iter": 1,
            f"{action}_width": self.working_width,
            f"{action}_height": self.working_height,
            f"{action}_n_samples": samples,
            f"{action}_scale": self.scale / 100,
            f"{action}_seed": self.seed,
            f"{action}_model": model,
            f"{action}_scheduler": self.scheduler,
            f"{action}_model_path": model_path,
            f"{action}_model_branch": model_branch,
            f"{action}_lora": self.available_lora(action),
            f"{action}_controlnet_conditioning_scale": self.controlnet_guidance_scale,
            f"generator_section": self.currentTabSection,
            f"width": self.working_width,
            f"height": self.working_height,
            "do_nsfw_filter": self.settings_manager.settings.nsfw_filter.get(),
            "model_base_path": self.model_base_path,
            "pos_x": 0,
            "pos_y": 0,
            "outpaint_box_rect": self.active_rect,
            "hf_token": self.settings_manager.settings.hf_api_key.get(),
            "enable_controlnet": self.enable_controlnet and self.controlnet is not None,
            "controlnet": self.controlnet,
            "deterministic_generation": self.deterministic,
            "deterministic_seed": False,
            "zeroshot": self.zeroshot
        }

        if action == "superresolution":
            options["original_image_width"] = self.canvas.current_active_image_data.image.width
            options["original_image_height"] = self.canvas.current_active_image_data.image.height

        if action in ["txt2img", "img2img", "outpaint", "depth2img", "pix2pix"]:
            options[f"{action}_strength"] = self.strength / 100.0

        if action == "pix2pix":
            options[f"pix2pix_image_guidance_scale"] = self.image_scale

        """
        Emitting generate_signal with options allows us to pass more options to the dict from
        modal windows such as the image interpolation window.
        """
        self.generate_signal.emit(options)

        memory_options = self.get_memory_options()

        data = {
            "action": action,
            "options": {
                **options,
                **extra_options,
                **memory_options
            }
        }
        # data = self.do_generate_data_injection(data)  # TODO: Extensions
        self.client.message = data

    """
    TODO: Extensions
    def do_generate_data_injection(self, data):
        for extension in self.settings_manager.settings.active_extensions.get():
            data = extension.generate_data_injection(data)
        return data
    """

    def tab_changed_callback(self, index):
        self.set_final_size_label()
        self.canvas.update()

    def set_default_values(self, section, tab):
        tab.prompt.setPlainText(self.prompt)
        tab.negative_prompt.setPlainText(self.negative_prompt)
        tab.steps_spinbox.setValue(self.steps)
        tab.scale_spinbox.setValue(self.scale / 100)
        if section == "pix2pix":
            val = self.settings_manager.settings.pix2pix_image_guidance_scale.get()
            tab.image_scale_spinbox.setValue(val / 100)
            if type(val) == float:
                val = int(val * 100)
            tab.image_scale_slider.setValue(val)
        try:
            tab.strength_spinbox.setValue(self.strength / 100)
        except:
            pass
        tab.seed.setText(str(self.seed))
        tab.samples_spinbox.setValue(self.samples)
        tab.model_dropdown.setCurrentText(self.model)
        try:
            tab.scheduler_dropdown.setCurrentText(self.scheduler)
        except RuntimeError:
            pass

    def load_template(self, template_name):
        try:
            return uic.loadUi(
                os.path.join("pyqt", f"{template_name}.ui"))
        except UIFileException:
            return None

    def clear_status_message(self):
        self.message_var.set("")



from airunner.canvas import Canvas


class CanvasMixin:
    def initialize(self):
        self.canvas = Canvas(parent=self)
        self.settings_manager.settings.show_grid.my_signal.connect(self.canvas.update)
        self.settings_manager.settings.snap_to_grid.my_signal.connect(self.canvas.update)
        self.settings_manager.settings.line_color.my_signal.connect(self.canvas.update_grid_pen)



# Form implementation generated from reading ui file '/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/widgets/prompt_builder.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(816, 405)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Form.sizePolicy().hasHeightForWidth())
        Form.setSizePolicy(sizePolicy)
        Form.setMinimumSize(QtCore.QSize(0, 405))
        font = QtGui.QFont()
        font.setPointSize(9)
        Form.setFont(font)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 0, 0, 9)
        self.gridLayout.setObjectName("gridLayout")
        self.line = QtWidgets.QFrame(parent=Form)
        self.line.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 1, 0, 1, 1)
        self.weights = QtWidgets.QGridLayout()
        self.weights.setObjectName("weights")
        self.label_5 = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        self.label_5.setFont(font)
        self.label_5.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight|QtCore.Qt.AlignmentFlag.AlignTrailing|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.label_5.setObjectName("label_5")
        self.weights.addWidget(self.label_5, 2, 4, 1, 1)
        self.label_4 = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        self.label_4.setFont(font)
        self.label_4.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeading|QtCore.Qt.AlignmentFlag.AlignLeft|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.label_4.setObjectName("label_4")
        self.weights.addWidget(self.label_4, 1, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        self.label_3.setFont(font)
        self.label_3.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight|QtCore.Qt.AlignmentFlag.AlignTrailing|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.label_3.setObjectName("label_3")
        self.weights.addWidget(self.label_3, 1, 4, 1, 1)
        self.negative_prompt_weight_distribution_slider = QtWidgets.QSlider(parent=Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.negative_prompt_weight_distribution_slider.sizePolicy().hasHeightForWidth())
        self.negative_prompt_weight_distribution_slider.setSizePolicy(sizePolicy)
        self.negative_prompt_weight_distribution_slider.setMinimumSize(QtCore.QSize(50, 0))
        self.negative_prompt_weight_distribution_slider.setMaximum(100)
        self.negative_prompt_weight_distribution_slider.setPageStep(1)
        self.negative_prompt_weight_distribution_slider.setProperty("value", 50)
        self.negative_prompt_weight_distribution_slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.negative_prompt_weight_distribution_slider.setTickPosition(QtWidgets.QSlider.TickPosition.NoTicks)
        self.negative_prompt_weight_distribution_slider.setTickInterval(1)
        self.negative_prompt_weight_distribution_slider.setObjectName("negative_prompt_weight_distribution_slider")
        self.weights.addWidget(self.negative_prompt_weight_distribution_slider, 2, 2, 1, 1)
        self.prompt_weight_distribution_slider = QtWidgets.QSlider(parent=Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.prompt_weight_distribution_slider.sizePolicy().hasHeightForWidth())
        self.prompt_weight_distribution_slider.setSizePolicy(sizePolicy)
        self.prompt_weight_distribution_slider.setMinimumSize(QtCore.QSize(50, 0))
        font = QtGui.QFont()
        font.setPointSize(10)
        self.prompt_weight_distribution_slider.setFont(font)
        self.prompt_weight_distribution_slider.setMaximum(100)
        self.prompt_weight_distribution_slider.setPageStep(1)
        self.prompt_weight_distribution_slider.setProperty("value", 50)
        self.prompt_weight_distribution_slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.prompt_weight_distribution_slider.setTickPosition(QtWidgets.QSlider.TickPosition.NoTicks)
        self.prompt_weight_distribution_slider.setTickInterval(1)
        self.prompt_weight_distribution_slider.setObjectName("prompt_weight_distribution_slider")
        self.weights.addWidget(self.prompt_weight_distribution_slider, 1, 2, 1, 1)
        self.auto_prompt_weight_label = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(True)
        self.auto_prompt_weight_label.setFont(font)
        self.auto_prompt_weight_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight|QtCore.Qt.AlignmentFlag.AlignTrailing|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.auto_prompt_weight_label.setObjectName("auto_prompt_weight_label")
        self.weights.addWidget(self.auto_prompt_weight_label, 1, 3, 1, 1)
        self.negative_text_prompt_weight_label = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(True)
        self.negative_text_prompt_weight_label.setFont(font)
        self.negative_text_prompt_weight_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeading|QtCore.Qt.AlignmentFlag.AlignLeft|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.negative_text_prompt_weight_label.setObjectName("negative_text_prompt_weight_label")
        self.weights.addWidget(self.negative_text_prompt_weight_label, 2, 1, 1, 1)
        self.label_6 = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        self.label_6.setFont(font)
        self.label_6.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeading|QtCore.Qt.AlignmentFlag.AlignLeft|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.label_6.setObjectName("label_6")
        self.weights.addWidget(self.label_6, 2, 0, 1, 1)
        self.label_7 = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        self.label_7.setFont(font)
        self.label_7.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.label_7.setObjectName("label_7")
        self.weights.addWidget(self.label_7, 0, 2, 1, 1)
        self.negative_auto_prompt_weight_label = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(True)
        self.negative_auto_prompt_weight_label.setFont(font)
        self.negative_auto_prompt_weight_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight|QtCore.Qt.AlignmentFlag.AlignTrailing|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.negative_auto_prompt_weight_label.setObjectName("negative_auto_prompt_weight_label")
        self.weights.addWidget(self.negative_auto_prompt_weight_label, 2, 3, 1, 1)
        self.text_prompt_weight_label = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(True)
        self.text_prompt_weight_label.setFont(font)
        self.text_prompt_weight_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeading|QtCore.Qt.AlignmentFlag.AlignLeft|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.text_prompt_weight_label.setObjectName("text_prompt_weight_label")
        self.weights.addWidget(self.text_prompt_weight_label, 1, 1, 1, 1)
        self.gridLayout.addLayout(self.weights, 2, 0, 1, 1)
        self.tabs = QtWidgets.QTabWidget(parent=Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabs.sizePolicy().hasHeightForWidth())
        self.tabs.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(9)
        self.tabs.setFont(font)
        self.tabs.setStyleSheet("")
        self.tabs.setTabBarAutoHide(True)
        self.tabs.setObjectName("tabs")
        self.advanced = QtWidgets.QWidget()
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.advanced.sizePolicy().hasHeightForWidth())
        self.advanced.setSizePolicy(sizePolicy)
        self.advanced.setMinimumSize(QtCore.QSize(0, 300))
        self.advanced.setObjectName("advanced")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.advanced)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.gridLayout_9 = QtWidgets.QGridLayout()
        self.gridLayout_9.setHorizontalSpacing(9)
        self.gridLayout_9.setObjectName("gridLayout_9")
        self.scrollArea = QtWidgets.QScrollArea(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scrollArea.sizePolicy().hasHeightForWidth())
        self.scrollArea.setSizePolicy(sizePolicy)
        self.scrollArea.setMinimumSize(QtCore.QSize(250, 120))
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scroll_grid = QtWidgets.QWidget()
        self.scroll_grid.setGeometry(QtCore.QRect(0, 0, 390, 214))
        self.scroll_grid.setObjectName("scroll_grid")
        self.scrollArea.setWidget(self.scroll_grid)
        self.gridLayout_9.addWidget(self.scrollArea, 0, 1, 1, 1)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_16 = QtWidgets.QLabel(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_16.sizePolicy().hasHeightForWidth())
        self.label_16.setSizePolicy(sizePolicy)
        self.label_16.setMaximumSize(QtCore.QSize(200, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        self.label_16.setFont(font)
        self.label_16.setObjectName("label_16")
        self.verticalLayout_3.addWidget(self.label_16)
        self.advanced_prompt_text = QtWidgets.QTextBrowser(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.advanced_prompt_text.sizePolicy().hasHeightForWidth())
        self.advanced_prompt_text.setSizePolicy(sizePolicy)
        self.advanced_prompt_text.setMinimumSize(QtCore.QSize(200, 0))
        self.advanced_prompt_text.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        self.advanced_prompt_text.setFont(font)
        self.advanced_prompt_text.setReadOnly(False)
        self.advanced_prompt_text.setObjectName("advanced_prompt_text")
        self.verticalLayout_3.addWidget(self.advanced_prompt_text)
        self.label_17 = QtWidgets.QLabel(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_17.sizePolicy().hasHeightForWidth())
        self.label_17.setSizePolicy(sizePolicy)
        self.label_17.setMaximumSize(QtCore.QSize(200, 16777215))
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(True)
        self.label_17.setFont(font)
        self.label_17.setObjectName("label_17")
        self.verticalLayout_3.addWidget(self.label_17)
        self.advanced_negative_prompt_text = QtWidgets.QTextBrowser(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.advanced_negative_prompt_text.sizePolicy().hasHeightForWidth())
        self.advanced_negative_prompt_text.setSizePolicy(sizePolicy)
        self.advanced_negative_prompt_text.setMinimumSize(QtCore.QSize(200, 0))
        self.advanced_negative_prompt_text.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        self.advanced_negative_prompt_text.setFont(font)
        self.advanced_negative_prompt_text.setReadOnly(False)
        self.advanced_negative_prompt_text.setObjectName("advanced_negative_prompt_text")
        self.verticalLayout_3.addWidget(self.advanced_negative_prompt_text)
        self.gridLayout_9.addLayout(self.verticalLayout_3, 0, 2, 1, 1)
        self.gridLayout_2.addLayout(self.gridLayout_9, 1, 0, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.randomize_values_button = QtWidgets.QPushButton(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.randomize_values_button.sizePolicy().hasHeightForWidth())
        self.randomize_values_button.setSizePolicy(sizePolicy)
        self.randomize_values_button.setMinimumSize(QtCore.QSize(20, 0))
        self.randomize_values_button.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        self.randomize_values_button.setFont(font)
        self.randomize_values_button.setObjectName("randomize_values_button")
        self.horizontalLayout_3.addWidget(self.randomize_values_button)
        self.values_to_random_button = QtWidgets.QPushButton(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.values_to_random_button.sizePolicy().hasHeightForWidth())
        self.values_to_random_button.setSizePolicy(sizePolicy)
        self.values_to_random_button.setMinimumSize(QtCore.QSize(20, 0))
        self.values_to_random_button.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        self.values_to_random_button.setFont(font)
        self.values_to_random_button.setObjectName("values_to_random_button")
        self.horizontalLayout_3.addWidget(self.values_to_random_button)
        self.reset_weights_button = QtWidgets.QPushButton(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.reset_weights_button.sizePolicy().hasHeightForWidth())
        self.reset_weights_button.setSizePolicy(sizePolicy)
        self.reset_weights_button.setMinimumSize(QtCore.QSize(20, 0))
        self.reset_weights_button.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        self.reset_weights_button.setFont(font)
        self.reset_weights_button.setObjectName("reset_weights_button")
        self.horizontalLayout_3.addWidget(self.reset_weights_button)
        self.clear_values_button = QtWidgets.QPushButton(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.clear_values_button.sizePolicy().hasHeightForWidth())
        self.clear_values_button.setSizePolicy(sizePolicy)
        self.clear_values_button.setMinimumSize(QtCore.QSize(20, 0))
        self.clear_values_button.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        self.clear_values_button.setFont(font)
        self.clear_values_button.setObjectName("clear_values_button")
        self.horizontalLayout_3.addWidget(self.clear_values_button)
        self.gridLayout_2.addLayout(self.horizontalLayout_3, 2, 0, 1, 1)
        self.gridLayout_10 = QtWidgets.QGridLayout()
        self.gridLayout_10.setObjectName("gridLayout_10")
        self.label_2 = QtWidgets.QLabel(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.gridLayout_10.addWidget(self.label_2, 0, 0, 1, 1)
        self.label = QtWidgets.QLabel(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.gridLayout_10.addWidget(self.label, 0, 1, 1, 1)
        self.advanced_style = QtWidgets.QComboBox(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.advanced_style.sizePolicy().hasHeightForWidth())
        self.advanced_style.setSizePolicy(sizePolicy)
        self.advanced_style.setMinimumSize(QtCore.QSize(80, 22))
        self.advanced_style.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(9)
        self.advanced_style.setFont(font)
        self.advanced_style.setObjectName("advanced_style")
        self.gridLayout_10.addWidget(self.advanced_style, 1, 2, 1, 1)
        self.advanced_prompt = QtWidgets.QComboBox(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.advanced_prompt.sizePolicy().hasHeightForWidth())
        self.advanced_prompt.setSizePolicy(sizePolicy)
        self.advanced_prompt.setMinimumSize(QtCore.QSize(80, 22))
        self.advanced_prompt.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(9)
        self.advanced_prompt.setFont(font)
        self.advanced_prompt.setObjectName("advanced_prompt")
        self.gridLayout_10.addWidget(self.advanced_prompt, 1, 1, 1, 1)
        self.advanced_category = QtWidgets.QComboBox(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.advanced_category.sizePolicy().hasHeightForWidth())
        self.advanced_category.setSizePolicy(sizePolicy)
        self.advanced_category.setMinimumSize(QtCore.QSize(80, 22))
        self.advanced_category.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(9)
        self.advanced_category.setFont(font)
        self.advanced_category.setObjectName("advanced_category")
        self.gridLayout_10.addWidget(self.advanced_category, 1, 0, 1, 1)
        self.label_11 = QtWidgets.QLabel(parent=self.advanced)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_11.sizePolicy().hasHeightForWidth())
        self.label_11.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        self.label_11.setFont(font)
        self.label_11.setObjectName("label_11")
        self.gridLayout_10.addWidget(self.label_11, 0, 2, 1, 1)
        self.advanced_radio = QtWidgets.QRadioButton(parent=self.advanced)
        font = QtGui.QFont()
        font.setPointSize(8)
        self.advanced_radio.setFont(font)
        self.advanced_radio.setObjectName("advanced_radio")
        self.gridLayout_10.addWidget(self.advanced_radio, 1, 3, 1, 1)
        self.basic_radio = QtWidgets.QRadioButton(parent=self.advanced)
        font = QtGui.QFont()
        font.setPointSize(8)
        self.basic_radio.setFont(font)
        self.basic_radio.setObjectName("basic_radio")
        self.gridLayout_10.addWidget(self.basic_radio, 0, 3, 1, 1)
        self.gridLayout_2.addLayout(self.gridLayout_10, 0, 0, 1, 1)
        self.tabs.addTab(self.advanced, "")
        self.gridLayout.addWidget(self.tabs, 0, 0, 1, 1)

        self.retranslateUi(Form)
        self.tabs.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_5.setText(_translate("Form", "Auto Neg Prompt"))
        self.label_4.setText(_translate("Form", "Text Prompt"))
        self.label_3.setText(_translate("Form", "Auto Prompt"))
        self.auto_prompt_weight_label.setText(_translate("Form", "0.50"))
        self.negative_text_prompt_weight_label.setText(_translate("Form", "0.5"))
        self.label_6.setText(_translate("Form", "Negative Prompt"))
        self.label_7.setText(_translate("Form", "Weight distribution"))
        self.negative_auto_prompt_weight_label.setText(_translate("Form", "0.5"))
        self.text_prompt_weight_label.setText(_translate("Form", "0.50"))
        self.label_16.setText(_translate("Form", "Prompt"))
        self.label_17.setText(_translate("Form", "Negative Prompt"))
        self.randomize_values_button.setText(_translate("Form", "Randomize"))
        self.values_to_random_button.setText(_translate("Form", "To Random"))
        self.reset_weights_button.setText(_translate("Form", "Reset Weights"))
        self.clear_values_button.setText(_translate("Form", "Clear"))
        self.label_2.setText(_translate("Form", "Category"))
        self.label.setText(_translate("Form", "Prompt"))
        self.label_11.setText(_translate("Form", "Style"))
        self.advanced_radio.setText(_translate("Form", "Advanced"))
        self.basic_radio.setText(_translate("Form", "Basic"))
        self.tabs.setTabText(self.tabs.indexOf(self.advanced), _translate("Form", "Advanced"))



from PyQt6.QtGui import QColor
from PyQt6.QtWidgets import QColorDialog


class BrushesMixin:
    def set_primary_color(self):
        # display a color picker and keep on top
        color_name:str = self.settings_manager.settings.primary_color.get()
        qcolor = QColor(color_name)
        color_dialog = QColorDialog(self)
        # color = self.layout().addWidget(color_dialog)
        color = color_dialog.getColor(qcolor, None, "Select Color")
        if color.isValid():
            self.settings_manager.settings.primary_color.set(color.name())



# Form implementation generated from reading ui file '/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/main_window.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1195, 880)
        MainWindow.setMinimumSize(QtCore.QSize(1195, 880))
        self.centralwidget = QtWidgets.QWidget(parent=MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.centralwidget.sizePolicy().hasHeightForWidth())
        self.centralwidget.setSizePolicy(sizePolicy)
        self.centralwidget.setMinimumSize(QtCore.QSize(1000, 512))
        self.centralwidget.setLayoutDirection(QtCore.Qt.LayoutDirection.LeftToRight)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setContentsMargins(9, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.tabWidget = QtWidgets.QTabWidget(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabWidget.sizePolicy().hasHeightForWidth())
        self.tabWidget.setSizePolicy(sizePolicy)
        self.tabWidget.setMinimumSize(QtCore.QSize(400, 0))
        self.tabWidget.setMaximumSize(QtCore.QSize(400, 16777215))
        self.tabWidget.setLayoutDirection(QtCore.Qt.LayoutDirection.LeftToRight)
        self.tabWidget.setStyleSheet("")
        self.tabWidget.setTabPosition(QtWidgets.QTabWidget.TabPosition.West)
        self.tabWidget.setTabShape(QtWidgets.QTabWidget.TabShape.Rounded)
        self.tabWidget.setElideMode(QtCore.Qt.TextElideMode.ElideNone)
        self.tabWidget.setDocumentMode(False)
        self.tabWidget.setTabsClosable(False)
        self.tabWidget.setMovable(True)
        self.tabWidget.setTabBarAutoHide(False)
        self.tabWidget.setObjectName("tabWidget")
        self.gridLayout.addWidget(self.tabWidget, 3, 0, 1, 1)
        self.right_toolbar = QtWidgets.QWidget(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.right_toolbar.sizePolicy().hasHeightForWidth())
        self.right_toolbar.setSizePolicy(sizePolicy)
        self.right_toolbar.setMinimumSize(QtCore.QSize(150, 30))
        font = QtGui.QFont()
        font.setItalic(False)
        self.right_toolbar.setFont(font)
        self.right_toolbar.setStyleSheet("border-top-color: rgba(191, 64, 64, 0);")
        self.right_toolbar.setObjectName("right_toolbar")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.right_toolbar)
        self.gridLayout_2.setContentsMargins(9, 0, 9, 0)
        self.gridLayout_2.setHorizontalSpacing(0)
        self.gridLayout_2.setVerticalSpacing(9)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.layers = QtWidgets.QScrollArea(parent=self.right_toolbar)
        self.layers.setFrameShadow(QtWidgets.QFrame.Shadow.Plain)
        self.layers.setWidgetResizable(True)
        self.layers.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeading|QtCore.Qt.AlignmentFlag.AlignLeft|QtCore.Qt.AlignmentFlag.AlignTop)
        self.layers.setObjectName("layers")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 239, 408))
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scrollAreaWidgetContents.sizePolicy().hasHeightForWidth())
        self.scrollAreaWidgetContents.setSizePolicy(sizePolicy)
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout.setObjectName("verticalLayout")
        self.layers.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout_2.addWidget(self.layers, 4, 0, 1, 1)
        self.horizontalWidget = QtWidgets.QWidget(parent=self.right_toolbar)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.horizontalWidget.sizePolicy().hasHeightForWidth())
        self.horizontalWidget.setSizePolicy(sizePolicy)
        self.horizontalWidget.setLayoutDirection(QtCore.Qt.LayoutDirection.LeftToRight)
        self.horizontalWidget.setObjectName("horizontalWidget")
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout(self.horizontalWidget)
        self.horizontalLayout_8.setContentsMargins(0, 0, -1, 0)
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.new_layer = QtWidgets.QPushButton(parent=self.horizontalWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.new_layer.sizePolicy().hasHeightForWidth())
        self.new_layer.setSizePolicy(sizePolicy)
        self.new_layer.setMinimumSize(QtCore.QSize(32, 32))
        self.new_layer.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.new_layer.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/file-add.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        icon.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/file-add.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.On)
        icon.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/file-add.png"), QtGui.QIcon.Mode.Disabled, QtGui.QIcon.State.Off)
        icon.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/file-add.png"), QtGui.QIcon.Mode.Disabled, QtGui.QIcon.State.On)
        self.new_layer.setIcon(icon)
        self.new_layer.setIconSize(QtCore.QSize(24, 24))
        self.new_layer.setObjectName("new_layer")
        self.horizontalLayout_8.addWidget(self.new_layer)
        self.layer_up_button = QtWidgets.QPushButton(parent=self.horizontalWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.layer_up_button.sizePolicy().hasHeightForWidth())
        self.layer_up_button.setSizePolicy(sizePolicy)
        self.layer_up_button.setMinimumSize(QtCore.QSize(32, 32))
        self.layer_up_button.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.layer_up_button.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/arrow-up.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.On)
        self.layer_up_button.setIcon(icon1)
        self.layer_up_button.setIconSize(QtCore.QSize(24, 24))
        self.layer_up_button.setObjectName("layer_up_button")
        self.horizontalLayout_8.addWidget(self.layer_up_button)
        self.layer_down_button = QtWidgets.QPushButton(parent=self.horizontalWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.layer_down_button.sizePolicy().hasHeightForWidth())
        self.layer_down_button.setSizePolicy(sizePolicy)
        self.layer_down_button.setMinimumSize(QtCore.QSize(32, 32))
        self.layer_down_button.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.layer_down_button.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/arrow-down.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.On)
        self.layer_down_button.setIcon(icon2)
        self.layer_down_button.setIconSize(QtCore.QSize(24, 24))
        self.layer_down_button.setObjectName("layer_down_button")
        self.horizontalLayout_8.addWidget(self.layer_down_button)
        self.delete_layer_button = QtWidgets.QPushButton(parent=self.horizontalWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.delete_layer_button.sizePolicy().hasHeightForWidth())
        self.delete_layer_button.setSizePolicy(sizePolicy)
        self.delete_layer_button.setMinimumSize(QtCore.QSize(32, 32))
        self.delete_layer_button.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.delete_layer_button.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/delete.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.On)
        self.delete_layer_button.setIcon(icon3)
        self.delete_layer_button.setIconSize(QtCore.QSize(24, 24))
        self.delete_layer_button.setObjectName("delete_layer_button")
        self.horizontalLayout_8.addWidget(self.delete_layer_button)
        self.gridLayout_2.addWidget(self.horizontalWidget, 5, 0, 1, 1)
        self.groupBox_4 = QtWidgets.QGroupBox(parent=self.right_toolbar)
        self.groupBox_4.setMinimumSize(QtCore.QSize(0, 25))
        self.groupBox_4.setObjectName("groupBox_4")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBox_4)
        self.gridLayout_3.setContentsMargins(9, 9, 9, 9)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.primary_color_button = QtWidgets.QToolButton(parent=self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.primary_color_button.sizePolicy().hasHeightForWidth())
        self.primary_color_button.setSizePolicy(sizePolicy)
        self.primary_color_button.setText("")
        self.primary_color_button.setObjectName("primary_color_button")
        self.gridLayout_3.addWidget(self.primary_color_button, 0, 0, 1, 1)
        self.gridLayout_2.addWidget(self.groupBox_4, 2, 0, 1, 1)
        self.groupBox_3 = QtWidgets.QGroupBox(parent=self.right_toolbar)
        self.groupBox_3.setMinimumSize(QtCore.QSize(0, 50))
        self.groupBox_3.setSizeIncrement(QtCore.QSize(0, 50))
        self.groupBox_3.setObjectName("groupBox_3")
        self.formLayout_2 = QtWidgets.QFormLayout(self.groupBox_3)
        self.formLayout_2.setObjectName("formLayout_2")
        self.brush_size_slider = QtWidgets.QSlider(parent=self.groupBox_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.brush_size_slider.sizePolicy().hasHeightForWidth())
        self.brush_size_slider.setSizePolicy(sizePolicy)
        self.brush_size_slider.setMinimumSize(QtCore.QSize(150, 0))
        self.brush_size_slider.setMinimum(1)
        self.brush_size_slider.setMaximum(100)
        self.brush_size_slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.brush_size_slider.setObjectName("brush_size_slider")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.ItemRole.LabelRole, self.brush_size_slider)
        self.brush_size_spinbox = QtWidgets.QSpinBox(parent=self.groupBox_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.brush_size_spinbox.sizePolicy().hasHeightForWidth())
        self.brush_size_spinbox.setSizePolicy(sizePolicy)
        self.brush_size_spinbox.setMaximumSize(QtCore.QSize(75, 16777215))
        self.brush_size_spinbox.setMinimum(1)
        self.brush_size_spinbox.setMaximum(100)
        self.brush_size_spinbox.setObjectName("brush_size_spinbox")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.ItemRole.FieldRole, self.brush_size_spinbox)
        self.gridLayout_2.addWidget(self.groupBox_3, 1, 0, 1, 1)
        self.groupBox = QtWidgets.QGroupBox(parent=self.right_toolbar)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setLayoutDirection(QtCore.Qt.LayoutDirection.LeftToRight)
        self.groupBox.setAutoFillBackground(False)
        self.groupBox.setObjectName("groupBox")
        self.formLayout = QtWidgets.QFormLayout(self.groupBox)
        self.formLayout.setObjectName("formLayout")
        self.label_2 = QtWidgets.QLabel(parent=self.groupBox)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(False)
        self.label_2.setFont(font)
        self.label_2.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_2)
        self.width_slider = QtWidgets.QSlider(parent=self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.width_slider.sizePolicy().hasHeightForWidth())
        self.width_slider.setSizePolicy(sizePolicy)
        self.width_slider.setMinimumSize(QtCore.QSize(150, 0))
        self.width_slider.setMinimum(8)
        self.width_slider.setMaximum(4096)
        self.width_slider.setSingleStep(8)
        self.width_slider.setPageStep(8)
        self.width_slider.setProperty("value", 8)
        self.width_slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.width_slider.setTickPosition(QtWidgets.QSlider.TickPosition.NoTicks)
        self.width_slider.setTickInterval(8)
        self.width_slider.setObjectName("width_slider")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.ItemRole.LabelRole, self.width_slider)
        self.width_spinbox = QtWidgets.QSpinBox(parent=self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.width_spinbox.sizePolicy().hasHeightForWidth())
        self.width_spinbox.setSizePolicy(sizePolicy)
        self.width_spinbox.setMinimumSize(QtCore.QSize(0, 0))
        self.width_spinbox.setMaximumSize(QtCore.QSize(75, 16777215))
        self.width_spinbox.setMinimum(8)
        self.width_spinbox.setMaximum(4096)
        self.width_spinbox.setSingleStep(8)
        self.width_spinbox.setProperty("value", 8)
        self.width_spinbox.setDisplayIntegerBase(10)
        self.width_spinbox.setObjectName("width_spinbox")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.ItemRole.FieldRole, self.width_spinbox)
        self.label = QtWidgets.QLabel(parent=self.groupBox)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(False)
        self.label.setFont(font)
        self.label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.label.setObjectName("label")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label)
        self.height_slider = QtWidgets.QSlider(parent=self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.height_slider.sizePolicy().hasHeightForWidth())
        self.height_slider.setSizePolicy(sizePolicy)
        self.height_slider.setMinimumSize(QtCore.QSize(150, 0))
        self.height_slider.setSizeIncrement(QtCore.QSize(0, 0))
        font = QtGui.QFont()
        font.setPointSize(11)
        self.height_slider.setFont(font)
        self.height_slider.setMinimum(1)
        self.height_slider.setMaximum(4096)
        self.height_slider.setSingleStep(8)
        self.height_slider.setPageStep(8)
        self.height_slider.setProperty("value", 8)
        self.height_slider.setSliderPosition(8)
        self.height_slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.height_slider.setTickInterval(8)
        self.height_slider.setObjectName("height_slider")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.ItemRole.LabelRole, self.height_slider)
        self.height_spinbox = QtWidgets.QSpinBox(parent=self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.height_spinbox.sizePolicy().hasHeightForWidth())
        self.height_spinbox.setSizePolicy(sizePolicy)
        self.height_spinbox.setMinimumSize(QtCore.QSize(0, 0))
        self.height_spinbox.setMaximumSize(QtCore.QSize(75, 16777215))
        font = QtGui.QFont()
        font.setPointSize(11)
        self.height_spinbox.setFont(font)
        self.height_spinbox.setMinimum(8)
        self.height_spinbox.setMaximum(4096)
        self.height_spinbox.setSingleStep(8)
        self.height_spinbox.setProperty("value", 8)
        self.height_spinbox.setDisplayIntegerBase(10)
        self.height_spinbox.setObjectName("height_spinbox")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.ItemRole.FieldRole, self.height_spinbox)
        self.gridLayout_2.addWidget(self.groupBox, 0, 0, 1, 1)
        self.gridLayout.addWidget(self.right_toolbar, 3, 2, 1, 1)
        self.frame_2 = QtWidgets.QFrame(parent=self.centralwidget)
        self.frame_2.setMinimumSize(QtCore.QSize(0, 30))
        self.frame_2.setBaseSize(QtCore.QSize(0, 15))
        self.frame_2.setObjectName("frame_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame_2)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.status_label = QtWidgets.QLabel(parent=self.frame_2)
        self.status_label.setStyleSheet("color: rgb(119, 118, 123);")
        self.status_label.setText("")
        self.status_label.setObjectName("status_label")
        self.horizontalLayout.addWidget(self.status_label)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.debug_label = QtWidgets.QLabel(parent=self.frame_2)
        self.debug_label.setText("")
        self.debug_label.setObjectName("debug_label")
        self.horizontalLayout.addWidget(self.debug_label)
        self.canvas_position = QtWidgets.QLabel(parent=self.frame_2)
        self.canvas_position.setObjectName("canvas_position")
        self.horizontalLayout.addWidget(self.canvas_position)
        self.gridLayout.addWidget(self.frame_2, 4, 1, 1, 1)
        self.canvas_container = QtWidgets.QFrame(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.canvas_container.sizePolicy().hasHeightForWidth())
        self.canvas_container.setSizePolicy(sizePolicy)
        self.canvas_container.setMinimumSize(QtCore.QSize(200, 200))
        self.canvas_container.setStyleSheet("")
        self.canvas_container.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.canvas_container.setFrameShadow(QtWidgets.QFrame.Shadow.Raised)
        self.canvas_container.setObjectName("canvas_container")
        self.gridLayout.addWidget(self.canvas_container, 3, 1, 1, 1)
        self.widget_2 = QtWidgets.QWidget(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widget_2.sizePolicy().hasHeightForWidth())
        self.widget_2.setSizePolicy(sizePolicy)
        self.widget_2.setMinimumSize(QtCore.QSize(200, 50))
        font = QtGui.QFont()
        font.setItalic(False)
        self.widget_2.setFont(font)
        self.widget_2.setStyleSheet("border-left-color: rgba(191, 64, 64, 0);")
        self.widget_2.setObjectName("widget_2")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.widget_2)
        self.horizontalLayout_3.setContentsMargins(0, 0, 9, 0)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.horizontalFrame = QtWidgets.QFrame(parent=self.widget_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.horizontalFrame.sizePolicy().hasHeightForWidth())
        self.horizontalFrame.setSizePolicy(sizePolicy)
        self.horizontalFrame.setObjectName("horizontalFrame")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.horizontalFrame)
        self.horizontalLayout_6.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetDefaultConstraint)
        self.horizontalLayout_6.setContentsMargins(1, -1, -1, -1)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.active_grid_area_button = QtWidgets.QPushButton(parent=self.horizontalFrame)
        self.active_grid_area_button.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.active_grid_area_button.sizePolicy().hasHeightForWidth())
        self.active_grid_area_button.setSizePolicy(sizePolicy)
        self.active_grid_area_button.setMinimumSize(QtCore.QSize(0, 0))
        self.active_grid_area_button.setMaximumSize(QtCore.QSize(60, 30))
        self.active_grid_area_button.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.active_grid_area_button.setText("")
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/stop.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.active_grid_area_button.setIcon(icon4)
        self.active_grid_area_button.setIconSize(QtCore.QSize(24, 24))
        self.active_grid_area_button.setCheckable(True)
        self.active_grid_area_button.setChecked(False)
        self.active_grid_area_button.setAutoExclusive(False)
        self.active_grid_area_button.setFlat(False)
        self.active_grid_area_button.setObjectName("active_grid_area_button")
        self.horizontalLayout_6.addWidget(self.active_grid_area_button)
        self.eraser_button = QtWidgets.QPushButton(parent=self.horizontalFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.eraser_button.sizePolicy().hasHeightForWidth())
        self.eraser_button.setSizePolicy(sizePolicy)
        self.eraser_button.setMinimumSize(QtCore.QSize(0, 0))
        self.eraser_button.setMaximumSize(QtCore.QSize(60, 30))
        self.eraser_button.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.eraser_button.setText("")
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/eraser.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.eraser_button.setIcon(icon5)
        self.eraser_button.setIconSize(QtCore.QSize(24, 24))
        self.eraser_button.setCheckable(True)
        self.eraser_button.setChecked(False)
        self.eraser_button.setFlat(False)
        self.eraser_button.setObjectName("eraser_button")
        self.horizontalLayout_6.addWidget(self.eraser_button)
        self.brush_button = QtWidgets.QPushButton(parent=self.horizontalFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.brush_button.sizePolicy().hasHeightForWidth())
        self.brush_button.setSizePolicy(sizePolicy)
        self.brush_button.setMinimumSize(QtCore.QSize(0, 0))
        self.brush_button.setMaximumSize(QtCore.QSize(60, 30))
        self.brush_button.setText("")
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/pen.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.brush_button.setIcon(icon6)
        self.brush_button.setIconSize(QtCore.QSize(24, 24))
        self.brush_button.setCheckable(True)
        self.brush_button.setObjectName("brush_button")
        self.horizontalLayout_6.addWidget(self.brush_button)
        self.move_button = QtWidgets.QPushButton(parent=self.horizontalFrame)
        self.move_button.setMaximumSize(QtCore.QSize(60, 30))
        self.move_button.setText("")
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/move.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.On)
        self.move_button.setIcon(icon7)
        self.move_button.setIconSize(QtCore.QSize(24, 24))
        self.move_button.setCheckable(True)
        self.move_button.setObjectName("move_button")
        self.horizontalLayout_6.addWidget(self.move_button)
        self.grid_button = QtWidgets.QPushButton(parent=self.horizontalFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.grid_button.sizePolicy().hasHeightForWidth())
        self.grid_button.setSizePolicy(sizePolicy)
        self.grid_button.setMinimumSize(QtCore.QSize(0, 0))
        self.grid_button.setMaximumSize(QtCore.QSize(60, 30))
        self.grid_button.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.grid_button.setText("")
        icon8 = QtGui.QIcon()
        icon8.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/grid.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.grid_button.setIcon(icon8)
        self.grid_button.setIconSize(QtCore.QSize(24, 24))
        self.grid_button.setCheckable(True)
        self.grid_button.setObjectName("grid_button")
        self.horizontalLayout_6.addWidget(self.grid_button)
        self.nsfw_button = QtWidgets.QPushButton(parent=self.horizontalFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.nsfw_button.sizePolicy().hasHeightForWidth())
        self.nsfw_button.setSizePolicy(sizePolicy)
        self.nsfw_button.setMinimumSize(QtCore.QSize(0, 0))
        self.nsfw_button.setMaximumSize(QtCore.QSize(60, 30))
        self.nsfw_button.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.nsfw_button.setText("")
        icon9 = QtGui.QIcon()
        icon9.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/underwear.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.nsfw_button.setIcon(icon9)
        self.nsfw_button.setIconSize(QtCore.QSize(24, 24))
        self.nsfw_button.setCheckable(True)
        self.nsfw_button.setChecked(False)
        self.nsfw_button.setFlat(False)
        self.nsfw_button.setObjectName("nsfw_button")
        self.horizontalLayout_6.addWidget(self.nsfw_button)
        self.focus_button = QtWidgets.QPushButton(parent=self.horizontalFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.focus_button.sizePolicy().hasHeightForWidth())
        self.focus_button.setSizePolicy(sizePolicy)
        self.focus_button.setMinimumSize(QtCore.QSize(0, 0))
        self.focus_button.setMaximumSize(QtCore.QSize(60, 30))
        self.focus_button.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.focus_button.setText("")
        icon10 = QtGui.QIcon()
        icon10.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/camera-focus.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.focus_button.setIcon(icon10)
        self.focus_button.setIconSize(QtCore.QSize(24, 24))
        self.focus_button.setObjectName("focus_button")
        self.horizontalLayout_6.addWidget(self.focus_button)
        self.wordballoon_button = QtWidgets.QPushButton(parent=self.horizontalFrame)
        self.wordballoon_button.setMaximumSize(QtCore.QSize(30, 30))
        self.wordballoon_button.setText("")
        self.wordballoon_button.setObjectName("wordballoon_button")
        self.horizontalLayout_6.addWidget(self.wordballoon_button)
        self.undo_button = QtWidgets.QPushButton(parent=self.horizontalFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.undo_button.sizePolicy().hasHeightForWidth())
        self.undo_button.setSizePolicy(sizePolicy)
        self.undo_button.setMinimumSize(QtCore.QSize(0, 0))
        self.undo_button.setMaximumSize(QtCore.QSize(60, 30))
        self.undo_button.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.undo_button.setText("")
        icon11 = QtGui.QIcon()
        icon11.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/undo.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.undo_button.setIcon(icon11)
        self.undo_button.setIconSize(QtCore.QSize(24, 24))
        self.undo_button.setObjectName("undo_button")
        self.horizontalLayout_6.addWidget(self.undo_button)
        self.redo_button = QtWidgets.QPushButton(parent=self.horizontalFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.redo_button.sizePolicy().hasHeightForWidth())
        self.redo_button.setSizePolicy(sizePolicy)
        self.redo_button.setMinimumSize(QtCore.QSize(0, 0))
        self.redo_button.setMaximumSize(QtCore.QSize(60, 30))
        self.redo_button.setStyleSheet("border-color: rgb(61, 56, 70);")
        self.redo_button.setText("")
        icon12 = QtGui.QIcon()
        icon12.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/redo.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.redo_button.setIcon(icon12)
        self.redo_button.setIconSize(QtCore.QSize(24, 24))
        self.redo_button.setObjectName("redo_button")
        self.horizontalLayout_6.addWidget(self.redo_button)
        self.horizontalLayout_3.addWidget(self.horizontalFrame)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.darkmode_button = QtWidgets.QPushButton(parent=self.widget_2)
        self.darkmode_button.setMinimumSize(QtCore.QSize(24, 24))
        self.darkmode_button.setText("")
        icon13 = QtGui.QIcon()
        icon13.addPixmap(QtGui.QPixmap("/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/../src/icons/weather-night.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.On)
        self.darkmode_button.setIcon(icon13)
        self.darkmode_button.setIconSize(QtCore.QSize(24, 24))
        self.darkmode_button.setObjectName("darkmode_button")
        self.horizontalLayout_3.addWidget(self.darkmode_button)
        self.gridLayout.addWidget(self.widget_2, 0, 0, 1, 3)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(parent=MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1195, 22))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(parent=self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuEdit = QtWidgets.QMenu(parent=self.menubar)
        self.menuEdit.setObjectName("menuEdit")
        self.menuSettings = QtWidgets.QMenu(parent=self.menubar)
        self.menuSettings.setObjectName("menuSettings")
        self.menuFilters = QtWidgets.QMenu(parent=self.menubar)
        self.menuFilters.setObjectName("menuFilters")
        self.menuBlur = QtWidgets.QMenu(parent=self.menuFilters)
        self.menuBlur.setObjectName("menuBlur")
        self.menuAbout = QtWidgets.QMenu(parent=self.menubar)
        self.menuAbout.setTearOffEnabled(False)
        self.menuAbout.setObjectName("menuAbout")
        self.menuModel_merge = QtWidgets.QMenu(parent=self.menubar)
        self.menuModel_merge.setObjectName("menuModel_merge")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(parent=MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionNew = QtGui.QAction(parent=MainWindow)
        self.actionNew.setObjectName("actionNew")
        self.actionImport = QtGui.QAction(parent=MainWindow)
        self.actionImport.setObjectName("actionImport")
        self.actionExport = QtGui.QAction(parent=MainWindow)
        self.actionExport.setObjectName("actionExport")
        self.actionCopy = QtGui.QAction(parent=MainWindow)
        self.actionCopy.setObjectName("actionCopy")
        self.actionPaste = QtGui.QAction(parent=MainWindow)
        self.actionPaste.setObjectName("actionPaste")
        self.actionUndo = QtGui.QAction(parent=MainWindow)
        self.actionUndo.setObjectName("actionUndo")
        self.actionRedo = QtGui.QAction(parent=MainWindow)
        self.actionRedo.setObjectName("actionRedo")
        self.actionGrid = QtGui.QAction(parent=MainWindow)
        self.actionGrid.setObjectName("actionGrid")
        self.actionPreferences = QtGui.QAction(parent=MainWindow)
        self.actionPreferences.setObjectName("actionPreferences")
        self.actionCanvas_color = QtGui.QAction(parent=MainWindow)
        self.actionCanvas_color.setObjectName("actionCanvas_color")
        self.actionReset_Settings = QtGui.QAction(parent=MainWindow)
        self.actionReset_Settings.setObjectName("actionReset_Settings")
        self.actionResize_on_Paste = QtGui.QAction(parent=MainWindow)
        self.actionResize_on_Paste.setCheckable(True)
        self.actionResize_on_Paste.setObjectName("actionResize_on_Paste")
        self.actionMemory = QtGui.QAction(parent=MainWindow)
        self.actionMemory.setObjectName("actionMemory")
        self.actionAbout = QtGui.QAction(parent=MainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionCheck_for_updates = QtGui.QAction(parent=MainWindow)
        self.actionCheck_for_updates.setObjectName("actionCheck_for_updates")
        self.actionOpen = QtGui.QAction(parent=MainWindow)
        self.actionOpen.setObjectName("actionOpen")
        self.actionView = QtGui.QAction(parent=MainWindow)
        self.actionView.setObjectName("actionView")
        self.actionAdvanced = QtGui.QAction(parent=MainWindow)
        self.actionAdvanced.setObjectName("actionAdvanced")
        self.actionGaussian_Blur = QtGui.QAction(parent=MainWindow)
        self.actionGaussian_Blur.setEnabled(True)
        self.actionGaussian_Blur.setObjectName("actionGaussian_Blur")
        self.actionBox_Blur = QtGui.QAction(parent=MainWindow)
        self.actionBox_Blur.setObjectName("actionBox_Blur")
        self.actionUnsharp_Mask = QtGui.QAction(parent=MainWindow)
        self.actionUnsharp_Mask.setObjectName("actionUnsharp_Mask")
        self.actionPixel_Art = QtGui.QAction(parent=MainWindow)
        self.actionPixel_Art.setVisible(True)
        self.actionPixel_Art.setObjectName("actionPixel_Art")
        self.actionInvert = QtGui.QAction(parent=MainWindow)
        self.actionInvert.setObjectName("actionInvert")
        self.actionSaturation = QtGui.QAction(parent=MainWindow)
        self.actionSaturation.setVisible(True)
        self.actionSaturation.setObjectName("actionSaturation")
        self.actionColor_Balance = QtGui.QAction(parent=MainWindow)
        self.actionColor_Balance.setEnabled(True)
        self.actionColor_Balance.setVisible(True)
        self.actionColor_Balance.setObjectName("actionColor_Balance")
        self.actionSave = QtGui.QAction(parent=MainWindow)
        self.actionSave.setObjectName("actionSave")
        self.actionLoad = QtGui.QAction(parent=MainWindow)
        self.actionLoad.setObjectName("actionLoad")
        self.actionQuit = QtGui.QAction(parent=MainWindow)
        self.actionQuit.setObjectName("actionQuit")
        self.actionPixel_Art_2 = QtGui.QAction(parent=MainWindow)
        self.actionPixel_Art_2.setObjectName("actionPixel_Art_2")
        self.actionBug_report = QtGui.QAction(parent=MainWindow)
        self.actionBug_report.setObjectName("actionBug_report")
        self.actionReport_vulnerability = QtGui.QAction(parent=MainWindow)
        self.actionReport_vulnerability.setObjectName("actionReport_vulnerability")
        self.actionDiscord = QtGui.QAction(parent=MainWindow)
        self.actionDiscord.setObjectName("actionDiscord")
        self.actionExtensions = QtGui.QAction(parent=MainWindow)
        self.actionExtensions.setObjectName("actionExtensions")
        self.actionShow_image_preview = QtGui.QAction(parent=MainWindow)
        self.actionShow_image_preview.setCheckable(True)
        self.actionShow_image_preview.setObjectName("actionShow_image_preview")
        self.actionImage_to_new_layer = QtGui.QAction(parent=MainWindow)
        self.actionImage_to_new_layer.setCheckable(True)
        self.actionImage_to_new_layer.setObjectName("actionImage_to_new_layer")
        self.actionAuto_export_images = QtGui.QAction(parent=MainWindow)
        self.actionAuto_export_images.setCheckable(True)
        self.actionAuto_export_images.setObjectName("actionAuto_export_images")
        self.actionImage_export_settings = QtGui.QAction(parent=MainWindow)
        self.actionImage_export_settings.setObjectName("actionImage_export_settings")
        self.actionCheck_for_latest_version_on_startup = QtGui.QAction(parent=MainWindow)
        self.actionCheck_for_latest_version_on_startup.setCheckable(True)
        self.actionCheck_for_latest_version_on_startup.setObjectName("actionCheck_for_latest_version_on_startup")
        self.actionModel_Merger = QtGui.QAction(parent=MainWindow)
        self.actionModel_Merger.setObjectName("actionModel_Merger")
        self.actionSaturation_Filter = QtGui.QAction(parent=MainWindow)
        self.actionSaturation_Filter.setObjectName("actionSaturation_Filter")
        self.actionColor_Balance_2 = QtGui.QAction(parent=MainWindow)
        self.actionColor_Balance_2.setObjectName("actionColor_Balance_2")
        self.actionRGB_Noise = QtGui.QAction(parent=MainWindow)
        self.actionRGB_Noise.setObjectName("actionRGB_Noise")
        self.actionGaussian_Blur_2 = QtGui.QAction(parent=MainWindow)
        self.actionGaussian_Blur_2.setObjectName("actionGaussian_Blur_2")
        self.actionBox_Blur_2 = QtGui.QAction(parent=MainWindow)
        self.actionBox_Blur_2.setObjectName("actionBox_Blur_2")
        self.actionRotate_90_clockwise = QtGui.QAction(parent=MainWindow)
        self.actionRotate_90_clockwise.setObjectName("actionRotate_90_clockwise")
        self.actionRotate_90_counter_clockwise = QtGui.QAction(parent=MainWindow)
        self.actionRotate_90_counter_clockwise.setObjectName("actionRotate_90_counter_clockwise")
        self.actionSave_prompt = QtGui.QAction(parent=MainWindow)
        self.actionSave_prompt.setObjectName("actionSave_prompt")
        self.actionPrompt_Browser = QtGui.QAction(parent=MainWindow)
        self.actionPrompt_Browser.setObjectName("actionPrompt_Browser")
        self.menuFile.addAction(self.actionNew)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSave)
        self.menuFile.addAction(self.actionLoad)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionExport)
        self.menuFile.addAction(self.actionImport)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSave_prompt)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionQuit)
        self.menuEdit.addAction(self.actionCopy)
        self.menuEdit.addAction(self.actionPaste)
        self.menuEdit.addAction(self.actionUndo)
        self.menuEdit.addAction(self.actionRedo)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionRotate_90_clockwise)
        self.menuEdit.addAction(self.actionRotate_90_counter_clockwise)
        self.menuSettings.addAction(self.actionPreferences)
        self.menuSettings.addSeparator()
        self.menuSettings.addAction(self.actionGrid)
        self.menuSettings.addAction(self.actionCanvas_color)
        self.menuSettings.addSeparator()
        self.menuSettings.addAction(self.actionCheck_for_latest_version_on_startup)
        self.menuSettings.addAction(self.actionResize_on_Paste)
        self.menuSettings.addAction(self.actionImage_to_new_layer)
        self.menuSettings.addSeparator()
        self.menuSettings.addAction(self.actionImage_export_settings)
        self.menuSettings.addSeparator()
        self.menuSettings.addAction(self.actionAdvanced)
        self.menuSettings.addSeparator()
        self.menuSettings.addAction(self.actionReset_Settings)
        self.menuBlur.addSeparator()
        self.menuBlur.addAction(self.actionGaussian_Blur_2)
        self.menuBlur.addAction(self.actionBox_Blur_2)
        self.menuFilters.addAction(self.menuBlur.menuAction())
        self.menuFilters.addSeparator()
        self.menuFilters.addAction(self.actionColor_Balance)
        self.menuFilters.addAction(self.actionSaturation_Filter)
        self.menuFilters.addAction(self.actionUnsharp_Mask)
        self.menuFilters.addAction(self.actionRGB_Noise)
        self.menuFilters.addAction(self.actionPixel_Art)
        self.menuFilters.addSeparator()
        self.menuFilters.addAction(self.actionInvert)
        self.menuAbout.addAction(self.actionAbout)
        self.menuAbout.addAction(self.actionBug_report)
        self.menuAbout.addAction(self.actionReport_vulnerability)
        self.menuAbout.addAction(self.actionDiscord)
        self.menuModel_merge.addAction(self.actionModel_Merger)
        self.menuModel_merge.addSeparator()
        self.menuModel_merge.addAction(self.actionPrompt_Browser)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuSettings.menuAction())
        self.menubar.addAction(self.menuModel_merge.menuAction())
        self.menubar.addAction(self.menuFilters.menuAction())
        self.menubar.addAction(self.menuAbout.menuAction())

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(-1)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "AI Runner"))
        self.new_layer.setToolTip(_translate("MainWindow", "Create a new layer"))
        self.groupBox_4.setTitle(_translate("MainWindow", "Brush color"))
        self.groupBox_3.setTitle(_translate("MainWindow", "Brush size"))
        self.groupBox.setTitle(_translate("MainWindow", "Active grid area"))
        self.label_2.setText(_translate("MainWindow", "Width"))
        self.label.setText(_translate("MainWindow", "Height"))
        self.canvas_position.setText(_translate("MainWindow", "0, 0"))
        self.active_grid_area_button.setToolTip(_translate("MainWindow", "Active Grid Area"))
        self.active_grid_area_button.setShortcut(_translate("MainWindow", "M"))
        self.eraser_button.setToolTip(_translate("MainWindow", "Eraser"))
        self.eraser_button.setShortcut(_translate("MainWindow", "E"))
        self.brush_button.setToolTip(_translate("MainWindow", "Pen"))
        self.brush_button.setStyleSheet(_translate("MainWindow", "border-color: rgb(61, 56, 70);"))
        self.brush_button.setShortcut(_translate("MainWindow", "B"))
        self.move_button.setStyleSheet(_translate("MainWindow", "border-color: rgb(61, 56, 70);"))
        self.grid_button.setToolTip(_translate("MainWindow", "Toggle Grid"))
        self.grid_button.setShortcut(_translate("MainWindow", "G"))
        self.nsfw_button.setToolTip(_translate("MainWindow", "Toggle NSFW Filter"))
        self.focus_button.setToolTip(_translate("MainWindow", "Recenter Canvas"))
        self.undo_button.setToolTip(_translate("MainWindow", "Undo"))
        self.redo_button.setToolTip(_translate("MainWindow", "Redo"))
        self.darkmode_button.setToolTip(_translate("MainWindow", "Toggle dark mode"))
        self.darkmode_button.setStyleSheet(_translate("MainWindow", "border-color: rgb(61, 56, 70);"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.menuEdit.setTitle(_translate("MainWindow", "Edit"))
        self.menuSettings.setTitle(_translate("MainWindow", "Settings"))
        self.menuFilters.setTitle(_translate("MainWindow", "Filters"))
        self.menuBlur.setTitle(_translate("MainWindow", "Blur"))
        self.menuAbout.setTitle(_translate("MainWindow", "Help"))
        self.menuModel_merge.setTitle(_translate("MainWindow", "Tools"))
        self.actionNew.setText(_translate("MainWindow", "New"))
        self.actionNew.setShortcut(_translate("MainWindow", "Ctrl+N"))
        self.actionImport.setText(_translate("MainWindow", "Import image"))
        self.actionImport.setShortcut(_translate("MainWindow", "Ctrl+I"))
        self.actionExport.setText(_translate("MainWindow", "Export image"))
        self.actionExport.setShortcut(_translate("MainWindow", "Ctrl+E"))
        self.actionCopy.setText(_translate("MainWindow", "Copy"))
        self.actionCopy.setShortcut(_translate("MainWindow", "Ctrl+C"))
        self.actionPaste.setText(_translate("MainWindow", "Paste"))
        self.actionPaste.setShortcut(_translate("MainWindow", "Ctrl+V"))
        self.actionUndo.setText(_translate("MainWindow", "Undo"))
        self.actionUndo.setShortcut(_translate("MainWindow", "Ctrl+Z"))
        self.actionRedo.setText(_translate("MainWindow", "Redo"))
        self.actionRedo.setShortcut(_translate("MainWindow", "Ctrl+Y"))
        self.actionGrid.setText(_translate("MainWindow", "Grid"))
        self.actionGrid.setToolTip(_translate("MainWindow", "Grid settings"))
        self.actionPreferences.setText(_translate("MainWindow", "Path Preferences"))
        self.actionPreferences.setToolTip(_translate("MainWindow", "Modify model path and more"))
        self.actionCanvas_color.setText(_translate("MainWindow", "Canvas color"))
        self.actionCanvas_color.setToolTip(_translate("MainWindow", "Change the color of the canvas"))
        self.actionReset_Settings.setText(_translate("MainWindow", "Reset Settings to Default"))
        self.actionReset_Settings.setToolTip(_translate("MainWindow", "Reset all settings to default"))
        self.actionResize_on_Paste.setText(_translate("MainWindow", "Resize on Import"))
        self.actionResize_on_Paste.setToolTip(_translate("MainWindow", "Resize the image to active grid area when importing"))
        self.actionMemory.setText(_translate("MainWindow", "Memory"))
        self.actionAbout.setText(_translate("MainWindow", "About"))
        self.actionCheck_for_updates.setText(_translate("MainWindow", "Check for updates"))
        self.actionOpen.setText(_translate("MainWindow", "Open"))
        self.actionView.setText(_translate("MainWindow", "View"))
        self.actionAdvanced.setText(_translate("MainWindow", "Memory Preferences"))
        self.actionAdvanced.setToolTip(_translate("MainWindow", "Change memory settings"))
        self.actionGaussian_Blur.setText(_translate("MainWindow", "Gaussian Blur"))
        self.actionBox_Blur.setText(_translate("MainWindow", "Box Blur"))
        self.actionUnsharp_Mask.setText(_translate("MainWindow", "Unsharp Mask"))
        self.actionPixel_Art.setText(_translate("MainWindow", "Pixel Art"))
        self.actionInvert.setText(_translate("MainWindow", "Invert"))
        self.actionSaturation.setText(_translate("MainWindow", "Saturation"))
        self.actionColor_Balance.setText(_translate("MainWindow", "Color Balance"))
        self.actionSave.setText(_translate("MainWindow", "Save"))
        self.actionSave.setShortcut(_translate("MainWindow", "Ctrl+S"))
        self.actionLoad.setText(_translate("MainWindow", "Open"))
        self.actionLoad.setShortcut(_translate("MainWindow", "Ctrl+O"))
        self.actionQuit.setText(_translate("MainWindow", "Quit"))
        self.actionQuit.setShortcut(_translate("MainWindow", "Ctrl+Q"))
        self.actionPixel_Art_2.setText(_translate("MainWindow", "Pixel Art"))
        self.actionBug_report.setText(_translate("MainWindow", "Bug report"))
        self.actionReport_vulnerability.setText(_translate("MainWindow", "Report vulnerability"))
        self.actionDiscord.setText(_translate("MainWindow", "Discord"))
        self.actionExtensions.setText(_translate("MainWindow", "Extensions"))
        self.actionExtensions.setToolTip(_translate("MainWindow", "Install, update and delete extensions"))
        self.actionShow_image_preview.setText(_translate("MainWindow", "Show image preview"))
        self.actionImage_to_new_layer.setText(_translate("MainWindow", "Image to new layer"))
        self.actionImage_to_new_layer.setToolTip(_translate("MainWindow", "Send generated images to new layer"))
        self.actionAuto_export_images.setText(_translate("MainWindow", "Auto export images"))
        self.actionAuto_export_images.setToolTip(_translate("MainWindow", "Automatically export newly generated images to a folder"))
        self.actionImage_export_settings.setText(_translate("MainWindow", "Import / export preferences"))
        self.actionCheck_for_latest_version_on_startup.setText(_translate("MainWindow", "Check for latest version on startup"))
        self.actionModel_Merger.setText(_translate("MainWindow", "Model Merger"))
        self.actionSaturation_Filter.setText(_translate("MainWindow", "Saturation"))
        self.actionColor_Balance_2.setText(_translate("MainWindow", "Color Balance"))
        self.actionRGB_Noise.setText(_translate("MainWindow", "RGB Noise"))
        self.actionGaussian_Blur_2.setText(_translate("MainWindow", "Gaussian Blur"))
        self.actionBox_Blur_2.setText(_translate("MainWindow", "Box Blur"))
        self.actionRotate_90_clockwise.setText(_translate("MainWindow", "Rotate 90° clockwise"))
        self.actionRotate_90_clockwise.setShortcut(_translate("MainWindow", "Ctrl+R"))
        self.actionRotate_90_counter_clockwise.setText(_translate("MainWindow", "Rotate 90° counter clockwise"))
        self.actionRotate_90_counter_clockwise.setShortcut(_translate("MainWindow", "Ctrl+Shift+R"))
        self.actionSave_prompt.setText(_translate("MainWindow", "Save prompt"))
        self.actionPrompt_Browser.setText(_translate("MainWindow", "Prompt Browser"))



import json
import queue
import time
from PyQt6 import QtCore
from PyQt6.QtCore import QThread
from aihandler.qtvar import BooleanVar
from aihandler.runner import SDRunner
import logging


class OfflineClient(QtCore.QObject):
    sd_runner = None
    app = None
    current_txt2img_model = None
    current_inpaint_model = None
    request_signal_status = QtCore.pyqtSignal(str)
    response_signal_status = QtCore.pyqtSignal(str)
    response_worker = None
    request_worker = None
    response_worker_thread = None
    request_worker_thread = None

    @property
    def message(self):
        """
        Does nothing. Only used for the setter.
        """
        return ""

    @message.setter
    def message(self, msg):
        """
        Set the message property
        """
        if msg == "cancel":
            self.logger.info("cancel message recieved")
            self.cancel()
        else:
            self.logger.info("Putting message in queue")
            self.queue.put(msg)

    @property
    def response(self):
        """
        Get the response from the server
        :return: response string
        """
        return ""

    @response.setter
    def response(self, msg):
        """
        Set the response
        :param msg:
        :return: None
        """
        self.res_queue.put(msg)

    def cancel(self):
        self.sd_runner.cancel()

    def __init__(self, **kwargs):
        super().__init__(
            parent=kwargs.get("parent", None)
        )
        self.app = kwargs.get("app", None)
        self.settings_manager = kwargs.get("settings_manager", None)
        self.quit_event = BooleanVar()
        self.queue = queue.Queue()
        self.res_queue = queue.Queue()
        self.quit_event.set(False)
        self.logger = logging.getLogger()
        self.message_var = kwargs.get("message_var")
        self.do_start()

    def do_start(self):
        # create a stable diffusion runner service
        # sd_runner_thread = self.start_thread(
        #     target=self.init_sd_runner,
        #     name="init stable diffusion runner"
        # )
        # sd_runner_thread.join()
        self.logger.info("Starting offline client")
        self.init_sd_runner()
        self.force_request_worker_reset()

    def init_sd_runner(self):
        # save sd_runner to disc and load from it next time
        # this is to avoid the overhead of creating a new sd_runner
        # every time we start the client
        self.logger.info("Initialzing SDRunner")
        self.sd_runner = SDRunner(
            app=self.app,
            message_var=self.message_var,
            settings_manager=self.settings_manager,
        )

    def handle_response(self, response):
        """
        Handle the response from the server
        :param response:
        :return: None
        """
        res = json.loads(response.decode("utf-8"))
        if "response" in res:
            self.response = response
        else:
            self.message = response

    def handle_error(self, error):
        self.logger.error(error)

    def callback(self, data):
        action = data.get("action")
        model = None
        model = data["options"][f"{data['action']}_model"]
        # on model change, reload the runner
        if (action in ("txt2img", "img2img") and self.current_txt2img_model != model) or (action in ("inpaint", "outpaint") and self.current_inpaint_model != model):
            do_reload = False
            if action in ("txt2img", "img2img"):
                if self.current_txt2img_model is not None:
                    do_reload = True
                self.current_txt2img_model = model
            elif action in ("inpaint", "outpaint"):
                if self.current_inpaint_model is not None:
                    do_reload = True
                self.current_inpaint_model = model
            if do_reload:
                # self.init_sd_runner()
                self.sd_runner.initialized = False
                self.sd_runner.reload_model = True

        if (action in ("txt2img", "img2img") and self.sd_runner.action in ("inpaint", "outpaint")) or \
            (action in ("inpaint", "outpaint") and self.sd_runner.action in ("txt2img", "img2img")):
            self.sd_runner.initialized = False

        self.sd_runner.generator_sample(data)

    def create_worker_thread(self):
        # start worker in a new thread using the self.worker method
        self.response_worker = ResponseWorker(client=self)
        self.request_worker = RequestWorker(client=self, callback=self.callback)
        self.response_worker_thread = QThread()
        self.response_worker_thread.started.connect(self.response_worker.startWork)
        self.request_worker_thread = QThread()
        self.request_worker_thread.started.connect(self.request_worker.startWork)
        self.response_worker_thread.start()
        self.request_worker_thread.start()
        self.response_worker.moveToThread(self.response_worker_thread)
        self.request_worker.moveToThread(self.request_worker_thread)
        self.response_worker.signalStatus.connect(self.request_signal_status)
        self.request_worker.signalStatus.connect(self.response_signal_status)

    def force_request_worker_reset(self):
        if self.request_worker_thread and self.request_worker_thread.isRunning():
            print('Terminating thread.')
            self.request_worker_thread.terminate()

            print('Waiting for thread termination.')
            self.request_worker_thread.wait()

            self.request_signal_status.emit('Idle.')

        if self.response_worker_thread and self.response_worker_thread.isRunning():
            print('Terminating thread.')
            self.response_worker_thread.terminate()

            print('Waiting for thread termination.')
            self.response_worker_thread.wait()

            self.response_signal_status.emit('Idle.')

        self.create_worker_thread()

    def force_request_worker_quit(self):
        if self.request_worker_thread.isRunning():
            self.request_worker_thread.terminate()
            self.request_worker_thread.wait()


class RequestWorker(QtCore.QObject):
    client: OfflineClient
    signalStatus = QtCore.pyqtSignal(str)

    def __init__(self, parent=None, client=None, callback=None):
        self.client = client
        super(self.__class__, self).__init__(None)
        self.callback = callback

    @QtCore.pyqtSlot()
    def startWork(self):
        while True:
            # check if we are connected to server
            if not self.client.queue.empty():
                try:
                    msg = self.client.queue.get(timeout=1)
                except queue.Empty:
                    msg = None
                if msg == "quit":
                    # self.parent.quit_event.set(True)
                    # break
                    pass
                self.callback(msg)
            time.sleep(0.01)


class ResponseWorker(QtCore.QObject):
    client: OfflineClient
    signalStatus = QtCore.pyqtSignal(str)

    def __init__(self, parent=None, client=None):
        self.client = client
        super(self.__class__, self).__init__(parent)

    @QtCore.pyqtSlot()
    def startWork(self):
        while True:
            try:
                msg = self.client.res_queue.get(timeout=1)
            except queue.Empty:
                msg = None
            if msg != "" and msg is not None:
                self.client.handle_response(msg)



import json
import random

from aihandler.prompt_parser import PromptParser
from airunner.build_prompt import BuildPrompt


class PromptData:
    data = None
    prompt_variables = {}
    categories = []
    genres = []
    colors = []
    styles = []
    variables = None
    weighted_variables = {}
    current_prompt = ""
    current_negative_prompt = ""
    text_prompt_weight = 0.0
    auto_prompt_weight = 1.0
    negative_text_prompt_weight = 0.0
    negative_auto_prompt_weight = 1.0
    seed = 42
    category = "Random"
    image_genre = "Random"
    image_color = "Random"
    image_style = "Random"
    advanced = False
    prompt_prefix = ""
    prompt_suffix = ""
    negative_prompt_prefix = ""
    negative_prompt_suffix = ""

    def __init__(self, file_name=""):
        self.prepare_data(file_name=file_name)

    def build_prompts(self, **kwargs):
        self.current_prompt = kwargs.get("prompt", self.current_prompt)
        self.current_negative_prompt = kwargs.get("negative_prompt", self.current_negative_prompt)
        self.text_prompt_weight = kwargs.get("text_prompt_weight", self.text_prompt_weight)
        self.auto_prompt_weight = kwargs.get("auto_prompt_weight", self.auto_prompt_weight)
        self.prompt_prefix = kwargs.get("prompt_prefix", self.prompt_prefix)
        self.prompt_suffix = kwargs.get("prompt_suffix", self.prompt_suffix)
        self.negative_prompt_prefix = kwargs.get("negative_prompt_prefix", self.negative_prompt_prefix)
        self.negative_prompt_suffix = kwargs.get("negative_prompt_suffix", self.negative_prompt_suffix)
        self.negative_text_prompt_weight = kwargs.get("negative_text_prompt_weight", self.negative_text_prompt_weight)
        self.negative_auto_prompt_weight = kwargs.get("negative_auto_prompt_weight", self.negative_auto_prompt_weight)
        self.weighted_variables = kwargs.get("weighted_variables", self.weighted_variables)
        self.seed = kwargs.get("seed", self.seed)
        self.category = kwargs.get("category", self.category)
        self.image_genre = kwargs.get("image_genre", self.image_genre)
        self.image_color = kwargs.get("image_color", self.image_color)
        self.image_style = kwargs.get("image_style", self.image_style)
        self.advanced = kwargs.get("advanced", self.advanced)

        category = self.category
        while category == "Random":
            category = random.choice(self.categories)

        image_genre = self.image_genre
        while image_genre == "Random":
            image_genre = random.choice(self.genres)

        image_color = self.image_color
        while image_color == "Random":
            image_color = random.choice(self.colors)

        image_style = self.image_style
        while image_style == "Random":
            image_style = random.choice(self.styles)

        if category in self.weighted_variables:
            weighted_variables = self.weighted_variables[category]
        else:
            weighted_variables = {}

        variables = self.filter_variables(category, image_genre, image_color, image_style)
        if not self.advanced:
            generated_prompt = self.get_basic_prompt_by_category(category, "standard")
        else:
            generated_prompt = BuildPrompt.build_prompt(
                conditionals=self.get_builder_by_category(category),
                vars=variables,
                category=category
            )

            # clean the generated_prompt before parsing
            generated_prompt = generated_prompt.strip()
            generated_prompt.replace("( ", "(")

        prefix = f"{self.prompt_prefix}, " if self.prompt_prefix != "" else ""
        suffix = f", {self.prompt_suffix}" if self.prompt_suffix != "" else ""
        generated_prompt = f"{prefix}($composition_genre, $composition_color, ($composition_style)++) {generated_prompt}{suffix}"

        # build the negative prompt
        prefix = f"{self.negative_prompt_prefix}, " if self.negative_prompt_prefix != "" else ""
        suffix = f", {self.negative_prompt_suffix}" if self.negative_prompt_suffix != "" else ""
        generated_negative_prompt = self.negative_prompt(category, image_style)
        generated_negative_prompt = f"{prefix} {generated_negative_prompt} {suffix}"

        # process the prompts
        prompt, negative_prompt = PromptParser.parse(
            prompt=self.current_prompt,
            negative_prompt=self.current_negative_prompt,
            generated_prompt=generated_prompt,
            generated_negative_prompt=generated_negative_prompt,
            text_weight=self.text_prompt_weight,
            auto_weight=self.auto_prompt_weight,
            negative_text_weight=self.negative_text_prompt_weight,
            negative_auto_weight=self.negative_auto_prompt_weight,
            variables=variables,
            weights=weighted_variables,
            seed=self.seed)

        return prompt, negative_prompt

    def available_prompts_by_category(self, category):
        return self.data["categories"][category]["prompts"].keys()

    def available_variables_by_category(self, category):
        return self.data["categories"][category]["variables"].keys()

    def variable_weights_by_category(self, category, variable):
        return self.data["categories"][category]["weights"][variable]

    def variable_values_by_category(self, category, variable):
        return self.data["categories"][category]["variables"][variable]

    def set_variable_values_by_category(self, category, variable, values):
        self.data["categories"][category]["variables"][variable] = values

    def prepare_variables(self, category, genre, color, style):
        variables = self.load_variables(category)
        variables = self.filter_composition_variables(genre, color, style, variables)
        return variables

    def get_basic_prompt_by_category(self, category, prompt_type):
        return self.data["categories"][category]["prompts"][prompt_type]

    def get_builder_by_category(self, category):
        return self.data["categories"][category]["builder"]

    def negative_prompt(self, category, image_style):
        negative_prompt_style_prefix = ""
        for style_category in self.style_categories():
            if image_style in self.styles_by_style_category(style_category):
                negative_prompt_style_prefix = self.negative_prompt_style_by_style_category(style_category)
                break
        generated_negative_prompt = self.negative_prompt_by_category(category)
        generated_negative_prompt = f"{negative_prompt_style_prefix} {generated_negative_prompt}"
        return generated_negative_prompt

    def negative_prompt_by_category(self, category):
        return self.data["categories"][category]["negative_prompt"]

    def negative_prompt_style_by_style_category(self, style_category):
        return self.data["styles"][style_category]["negative_prompt"]

    def style_categories(self):
        return self.data["styles"].keys()

    def styles_by_style_category(self, style_category):
        return self.data["styles"][style_category]["styles"]

    def filter_variables(self, category, genre, color, style):
        variables = self.prepare_variables(category, genre, color, style)

        weighted_variables = self.weighted_variables
        if category in weighted_variables:
            weighted_variables = weighted_variables[category]
            for variable in weighted_variables.keys():
                if variable in variables:
                    value = weighted_variables[variable]["value"]
                    if value == "":
                        del variables[variable]
                    elif value != "Random":
                        variables[variable] = [value]
        return variables

    def load_variables(self, category):
        """
        Loads the variables for the selected category.
        :return:
        """
        return {
            **self.prompt_variables.copy(),
            **self.data["categories"][category]["variables"].copy()
        }

    def filter_composition_variables(self, genre, color, style, variables):
        if genre not in ["", "Random"]:
            variables["composition_genre"] = [genre]
        if color not in ["", "Random"]:
            variables["composition_color"] = [color]
        if style not in ["", "Random"]:
            variables["composition_style"] = [style]
        return variables

    def prepare_genres(self):
        self.genres = self.prompt_variables["composition_genre"]
        self.genres.sort()
        self.genres = ["Random"] + self.prompt_variables["composition_genre"]

    def prepare_colors(self):
        self.colors = self.prompt_variables["composition_color"]
        self.colors.sort()
        self.colors = ["Random"] + self.prompt_variables["composition_color"]

    def prepare_styles(self):
        self.styles = self.prompt_variables["composition_style"]
        self.styles.sort()
        self.styles = ["Random"] + self.prompt_variables["composition_style"]

    def prepare_categories(self):
        self.categories = self.prompt_variables["composition_category"]
        self.categories.sort()
        self.categories = ["Random"] + self.prompt_variables["composition_category"]

    def prepare_data(self, file_name):
        self.load_data(file_name)
        self.prompt_variables = self.data["extra_variables"]
        self.prompt_variables["age"] = [str(n) for n in range(18, 100)]
        styles = []
        for style_category in self.data["styles"].keys():
            styles += self.data["styles"][style_category]["styles"]
        categories = list(self.data["categories"].keys())
        self.prompt_variables["composition_style"] = styles
        self.prompt_variables["composition_category"] = categories
        self.prepare_genres()
        self.prepare_styles()
        self.prepare_colors()
        self.prepare_categories()

    def load_data(self, file_name):
        file = f"data/{file_name}.json"
        with open(file, "r") as f:
            data = json.load(f)
            self.data = data



# Form implementation generated from reading ui file '/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/model_merger_model.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(430, 200)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Form.sizePolicy().hasHeightForWidth())
        Form.setSizePolicy(sizePolicy)
        Form.setMinimumSize(QtCore.QSize(0, 100))
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.models = QtWidgets.QComboBox(parent=Form)
        self.models.setObjectName("models")
        self.gridLayout.addWidget(self.models, 1, 0, 1, 1)
        self.model_label = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setBold(True)
        self.model_label.setFont(font)
        self.model_label.setObjectName("model_label")
        self.gridLayout.addWidget(self.model_label, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.model_weight_slider = QtWidgets.QSlider(parent=Form)
        self.model_weight_slider.setMaximum(100)
        self.model_weight_slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.model_weight_slider.setObjectName("model_weight_slider")
        self.horizontalLayout.addWidget(self.model_weight_slider)
        self.model_weight_spinbox = QtWidgets.QDoubleSpinBox(parent=Form)
        self.model_weight_spinbox.setMaximum(1.0)
        self.model_weight_spinbox.setObjectName("model_weight_spinbox")
        self.horizontalLayout.addWidget(self.model_weight_spinbox)
        self.gridLayout.addLayout(self.horizontalLayout, 2, 0, 1, 1)
        self.model_delete_button = QtWidgets.QPushButton(parent=Form)
        self.model_delete_button.setObjectName("model_delete_button")
        self.gridLayout.addWidget(self.model_delete_button, 3, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.gridLayout.addItem(spacerItem, 4, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.model_label.setText(_translate("Form", "Model"))
        self.model_delete_button.setText(_translate("Form", "Remove Model"))



# Form implementation generated from reading ui file '/home/joe/Projects/imagetopixel/airunner/src/airunner/pyqt/widgets/prompt_builder_variable_widget.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(200, 57)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Form.sizePolicy().hasHeightForWidth())
        Form.setSizePolicy(sizePolicy)
        Form.setMinimumSize(QtCore.QSize(200, 57))
        Form.setMaximumSize(QtCore.QSize(16777215, 57))
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setHorizontalSpacing(6)
        self.gridLayout.setVerticalSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.combobox = QtWidgets.QComboBox(parent=Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.combobox.sizePolicy().hasHeightForWidth())
        self.combobox.setSizePolicy(sizePolicy)
        self.combobox.setMinimumSize(QtCore.QSize(0, 0))
        self.combobox.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(False)
        self.combobox.setFont(font)
        self.combobox.setSizeAdjustPolicy(QtWidgets.QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.combobox.setMinimumContentsLength(10)
        self.combobox.setObjectName("combobox")
        self.gridLayout.addWidget(self.combobox, 1, 1, 1, 1)
        self.spinbox = QtWidgets.QDoubleSpinBox(parent=Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.spinbox.sizePolicy().hasHeightForWidth())
        self.spinbox.setSizePolicy(sizePolicy)
        self.spinbox.setMinimumSize(QtCore.QSize(20, 0))
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(False)
        self.spinbox.setFont(font)
        self.spinbox.setMaximum(2.0)
        self.spinbox.setSingleStep(0.01)
        self.spinbox.setProperty("value", 1.0)
        self.spinbox.setObjectName("spinbox")
        self.gridLayout.addWidget(self.spinbox, 1, 2, 1, 1)
        self.label = QtWidgets.QLabel(parent=Form)
        font = QtGui.QFont()
        font.setPointSize(8)
        font.setBold(True)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 1, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.spinbox.setToolTip(_translate("Form", "Weight"))
        self.label.setText(_translate("Form", "TextLabel"))



class History:
    event_history = []
    undone_history = []

    def add_event(self, data: dict):
        self.event_history.append(data)
        self.undone_history = []

    def clear(self):
        self.event_history = []
        self.undone_history = []



from PyQt6.QtCore import QPointF
from airunner.balloon import Balloon


class ComicMixin:
    def initialize(self):
        self.wordballoon_button.clicked.connect(self.word_balloon_button_clicked)

    def word_balloon_button_clicked(self):
        """
        Create and add a word balloon to the canvas.
        :return:
        """
        # create a word balloon
        word_balloon = Balloon()
        word_balloon.setGeometry(100, 100, 200, 100)
        word_balloon.set_tail_pos(QPointF(50, 100))
        # add the widget to the canvas
        self.history.add_event({
            "event": "add_widget",
            "layer_index": self.canvas.current_layer_index,
            "widgets": self.canvas.current_layer.widgets.copy(),
        })
        self.canvas.current_layer.widgets.append(word_balloon)
        self.show_layers()
        self.canvas.update()
